/* This tool handle the AppIcons and MiniWindows
   when GSUseIconManager is YES 

   Copyright (C) 2009 Free Software Foundation, Inc.

   Written by:  German Arias <german@xelalug.org>
   Created: December 2009

   This file is part of the GNUstep Project

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 3
   of the License, or (at your option) any later version.
    
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public  
   License along with this library; see the file COPYING.
   If not, see <http://www.gnu.org/licenses/> or write to the 
   Free Software Foundation, 51 Franklin Street, Fifth Floor, 
   Boston, MA 02110-1301, USA.
*/

#import <Foundation/Foundation.h>
#import <AppKit/AppKit.h>

/* This use the fallowing variables

originX = the X coordinate of the left lower corner
originY = the Y coordinate of the left lower corner
orderSignX = if is 1 the icons are added to the right
             if is -1 the icons are added to the left
             if is 0 there isn't change in X coordinate
orderSignY = if is 1 the icons are added to upwards
             if is -1 the icons are added to downwards
             if is 0 there isn't change in Y coordinate   

the array

icons = contains all icons, even the empty spaces between these

and a dictionary

windowsInfo = contains information about what processes have icons */

@interface GSIconManager : NSObject
{
  unsigned int originX, originY, edge;
  int orderSignX, orderSignY;
  NSMutableArray *icons;
  NSMutableDictionary *windowsInfo;
}
- (NSRect) setWindow: (unsigned int)aWindowNumber appProcessId: (int)aProcessId;
- (void) removeWindow: (unsigned int)aWindowNumber;
- (NSSize) getSizeWindow;
- (void) setPreferences: (NSArray *)preferences;
@end

@implementation GSIconManager
- (id) init
{
  self = [super init];
  NSTimer *timer;
  NSConnection *conn;
  icons = [NSMutableArray array];
  [icons retain];
  windowsInfo = [NSMutableDictionary dictionary];
  [windowsInfo retain];

  NSString *path = GSDefaultsRootForUser(NSUserName());
  path = [path stringByAppendingString: @"/gsim.plist"];
  NSArray *userPrefs = [NSArray arrayWithContentsOfFile: path];

  if ([userPrefs count] == 5)
    {
      NSLog(@"Reading user preferences...");
      orderSignX = [[[userPrefs objectAtIndex: 0] description] intValue];
      orderSignY = [[[userPrefs objectAtIndex: 1] description] intValue];
      originX = [[[userPrefs objectAtIndex: 2] description] intValue];    
      originY = [[[userPrefs objectAtIndex: 3] description] intValue];
      edge = [[[userPrefs objectAtIndex: 4] description] intValue];
    }
  else
    {
      NSLog(@"Can't read preferences, writing new file...");
      NSString *newFile = [NSString stringWithString: @"(0, 1, 0, 50, 32)"];
      [newFile writeToFile: path atomically: YES];
      orderSignX = 0;
      orderSignY = 1;
      originX = 0;
      originY = 50;
      edge = 32;    
    } 

  timer = [NSTimer scheduledTimerWithTimeInterval:30 
	                        	   target:self 
                                         selector:@selector(checkApps:) 
                                         userInfo:nil 
                                          repeats:YES] ;

  /* Get the default connection */
  conn = [NSConnection defaultConnection];

  /* Make the gsim available to other processes */
  [conn setRootObject: self];

  if (![conn registerName: @"GSIconManager"]) 
    {
      NSLog (@"Could not register us as GSIconManager");
      exit (1);
    }
  
  NSLog (@"Server registered - waiting for connections...");

  return self;
}

/* This method get an icon's window and give a rect for this.
If there isn't any icon, then this is placed at first position.
If there are icons, and there isn't any space available between
these, the new icon is positioned after the last icon. If there
is an available space between icons this is positioned in the 
first position available */

- (NSRect) setWindow: (unsigned int)aWindowNumber appProcessId: (int)aProcessId
{
  NSUInteger numWin;
  NSRect frame;
  NSNumber *winNum = [NSNumber numberWithUnsignedInt: aWindowNumber];

  if (![icons containsObject: [NSNull null]])
    {
      [icons addObject: winNum];
      numWin = [icons count] - 1;
    }
  else
    {
      numWin = [icons indexOfObject: [NSNull null]];
      [icons replaceObjectAtIndex: numWin withObject: winNum];
    }

  [windowsInfo setObject: [NSNumber numberWithInt: aProcessId] forKey: winNum];

  frame = NSMakeRect(originX + edge*orderSignX*numWin,
		     originY + edge*orderSignY*numWin,
		     edge,
		     edge);

  return frame;
}

/* This method remove an icon from the IconManager, and the 
position of this icon become available to others icons. 
If the icon that will be removed is on the last position,
this is deleted on the array icons. If not, the position
become available with an [NSNull null] object */

- (void) removeWindow: (unsigned int)aWindowNumber
{
  NSNumber *winNum = [NSNumber numberWithUnsignedInt: aWindowNumber];

  if ([icons containsObject: winNum])
    {
      NSUInteger ind = [icons indexOfObject: winNum];

      if ((ind + 1) == [icons count])
	{
	  [icons removeLastObject];
	}
      else
	{
	  [icons replaceObjectAtIndex: ind withObject: [NSNull null]];
	}

      [windowsInfo removeObjectForKey: winNum];
    }
}

/* This method return the size for an icon */

- (NSSize) getSizeWindow
{
  NSSize size = NSMakeSize(edge, edge);
  return size;
}

/* This method remove the icons of apps died */

- (void) checkApps: (NSTimer *)timer
{
  NSNumber *check, *checkId;
  NSDictionary *appInfo;
  NSEnumerator *counter;
  NSDictionary *storedWindows = AUTORELEASE([[NSDictionary alloc] initWithDictionary: windowsInfo 
								           copyItems: YES]);

  NSArray *workspaceApps = [[NSWorkspace sharedWorkspace] launchedApplications];
  NSArray *runningApps = AUTORELEASE([[NSArray alloc] initWithArray: workspaceApps
                                                          copyItems: YES]);

  NSMutableArray *runningIds = [NSMutableArray array];

  counter = [runningApps objectEnumerator];

  while ((appInfo = [counter nextObject]))
    {
      [runningIds addObject: [appInfo objectForKey: @"NSApplicationProcessIdentifier"]]; 
    }

  counter = [storedWindows keyEnumerator];

  while ((check = [counter nextObject]))
    {
      checkId = [storedWindows objectForKey: check];

      if (![runningIds containsObject: checkId])
	{
	  [self removeWindow: [check unsignedIntValue]]; 
	}
    }
}

/* This method set the user preferences */

- (void) setPreferences: (NSArray *)preferences
{
  orderSignX = [[[preferences objectAtIndex: 0] description] intValue];
  orderSignY = [[[preferences objectAtIndex: 1] description] intValue];
  originX = [[[preferences objectAtIndex: 2] description] intValue];
  originY = [[[preferences objectAtIndex: 3] description] intValue];
  edge = [[[preferences objectAtIndex: 4] description] intValue];  
}

- (void) dealloc
{
  RELEASE(icons);
  RELEASE(windowsInfo);
  [super dealloc];
}
@end


int 
main (void)
{
  BOOL subtask = YES;
  GSIconManager *GSIM;

  CREATE_AUTORELEASE_POOL(pool);

  NSProcessInfo *pInfo = [NSProcessInfo processInfo];
  NSMutableArray *args = AUTORELEASE([[pInfo arguments] mutableCopy]);

    if ([[pInfo arguments] containsObject: @"--daemon"] == YES)
    {
      subtask = NO;
    }

  if (subtask)
    {
      NSFileHandle	*null;
      NSTask		*t;
      
      t = [NSTask new];
      NS_DURING
	{
	  [args removeObjectAtIndex: 0];
	  [args addObject: @"--daemon"];
	  [t setLaunchPath: [[NSBundle mainBundle] executablePath]];
	  [t setArguments: args];
	  [t setEnvironment: [pInfo environment]];
	  null = [NSFileHandle fileHandleWithNullDevice];
	  [t setStandardInput: null];
	  [t setStandardOutput: null];
	  [t setStandardError: null];
	  [t launch];
	  DESTROY(t);
	}
      NS_HANDLER
	{
	  DESTROY(t);
	}
      NS_ENDHANDLER
	exit(EXIT_FAILURE);
    }
  RELEASE(pool);
  
  {
    CREATE_AUTORELEASE_POOL(pool);
    
    /* Make the tool gsim */
    GSIM = [GSIconManager new];
    
    /* Close standard input, output, and error to run as daemon */
    [[NSFileHandle fileHandleWithStandardInput] closeFile];
    [[NSFileHandle fileHandleWithStandardOutput] closeFile];
    [[NSFileHandle fileHandleWithStandardError] closeFile];
    
    RELEASE(pool);
  }
  
  if (GSIM != nil)
    {
      CREATE_AUTORELEASE_POOL(pool);
      [[NSRunLoop currentRunLoop] run];
      RELEASE(pool);
    }
  exit(EXIT_SUCCESS);
} 
